#!/usr/bin/env python3

import argparse

from importlib.metadata import version
from sys import exit

from Bio import SeqIO

NUCS = list("atgc")

BICODON_HEADERS = [
    "Division",
    "Assembly",
    "Taxid",
    "Species",
    "Organelle",
    "Translation Table",
    "# CDS",
    "# Codon Pairs",
]
CODON_HEADERS = [
    "Division",
    "Assembly",
    "Taxid",
    "Species",
    "Organelle",
    "Translation Table",
    "# CDS",
    "# Codons",
]
HEADER_COUNT = 8


def __main__():
    parser = argparse.ArgumentParser(
        description="Get Codon and Bicodon frequency from FASTA files"
    )
    parser.add_argument("--seq", default="in.fasta", help="FASTA nucleotide sequences")
    parser.add_argument("--taxid", default="9606", help="tax id")
    parser.add_argument("--organelle", default="genomic", help="organelle")
    parser.add_argument("--codon_out", default=None, help="Path to codon usage output")
    parser.add_argument(
        "--bicodon_out", default=None, help="Path to bicodon usage output"
    )
    parser.add_argument("--division", default="custom", help="division")
    parser.add_argument("--assembly", default="hg38", help="assembly")
    parser.add_argument("--species", default="Homo sapiens", help="species")
    parser.add_argument("--translation_table", default="1", help="translation_table")
    parser.add_argument(
        "--version", action="store_true", help="Report version and exit"
    )
    args = parser.parse_args()

    if args.version:
        print("nAltORFs {}".format(version("naltorfs")))
        exit()

    codons = {}
    for i in NUCS:
        for j in NUCS:
            for k in NUCS:
                codons[i + j + k] = 0

    bicodons = {}
    for i in codons.keys():
        for j in codons.keys():
            bicodons[i + j] = 0

    for seq_count, seq_record in enumerate(SeqIO.parse(args.seq, "fasta"), 1):
        seq = str(seq_record.seq).lower()
        n = 3
        for i in range(0, len(seq), n):
            codon = seq[i : i + n]
            codons[codon] = codons[codon] + 1
        for i in range(0, len(seq) - n, n):
            bicodon = seq[i : i + 2 * n]
            bicodons[bicodon] = bicodons[bicodon] + 1

    def write_counts(count_dict, headers, filename):
        with open(filename, "w") as fh:
            keys = list(count_dict.keys())
            keys.sort()
            header = headers + keys
            fh.write("%s\n" % "\t".join(header))
            fh.write(
                "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s"
                % (
                    args.division,
                    args.assembly,
                    args.taxid,
                    args.species,
                    args.organelle,
                    args.translation_table,
                    seq_count,
                    sum(count_dict.values()),
                )
            )
            for key in keys:
                fh.write("\t%i" % count_dict[key])
            fh.write("\n")

    write_counts(codons, CODON_HEADERS, args.codon_out)
    write_counts(bicodons, BICODON_HEADERS, args.bicodon_out)


if __name__ == "__main__":
    __main__()



#!/usr/bin/env python3

import argparse

from importlib.metadata import version
from sys import exit

from Bio.Seq import translate

# headers = "Division Assembly    Taxid   Species Organelle   Translation Table   # CDS   # Codon Pairs"
HEADER_COUNT = 8


def __main__():
    parser = argparse.ArgumentParser(
        description="Get Codon frequency from bicodon frequency (codon pairs)"
    )
    parser.add_argument(
        "--bicodons", default="o586358-Refseq_Bicod.tsv", help="Bicondon rates input"
    )
    parser.add_argument("--taxid", default="9606", help="tax id")
    parser.add_argument("--organelle", default="genomic", help="tax id")
    parser.add_argument("--out", default=None, help="Path to output")
    parser.add_argument("--aa_out", default=None, help="Path to AA output")
    parser.add_argument(
        "--version", action="store_true", help="Report version and exit"
    )
    args = parser.parse_args()

    if args.version:
        print("nAltORFs {}".format(version("naltorfs")))
        exit()

    codons1 = {}
    codons2 = {}
    aa1 = {}
    aa2 = {}
    with open(args.bicodons, "r") as fh:
        header = fh.readline().strip().split("\t")
        taxid_index = header.index("Taxid")
        organelle_index = header.index("Organelle")
        translation_table_index = header.index("Translation Table")
        # codon_pairs_index = header.index('# Codon Pairs')
        bicodons = header[HEADER_COUNT:]
        print("Using %i codon pairs" % (len(bicodons)))
        # TODO: should we confirm all codon pairs are actually listed in header?
        for line in fh:
            line = line.strip().split("\t")
            if (
                line[taxid_index] == args.taxid
                and line[organelle_index] == args.organelle
            ):
                bicodon_counts = line[HEADER_COUNT:]
                assert len(bicodons) == len(
                    bicodon_counts
                ), "Bicodons and bicodons counts are not same length"
                translation_table = int(line[translation_table_index])
                if translation_table < 1:
                    print(
                        "Translation table was reported as %i, which is invalid. Assuming standard code (1)."
                        % (translation_table)
                    )
                    translation_table = 1
                for bicodon, count in zip(bicodons, bicodon_counts):
                    codon = bicodon[1:4]
                    count = int(count)
                    if codon not in codons1:
                        codons1[codon] = []
                    codons1[codon].append(count)
                    aa = translate(codon, table=translation_table)
                    if aa not in aa1:
                        aa1[aa] = 0
                    aa1[aa] = aa1[aa] + count
                    codon = bicodon[2:5]
                    if codon not in codons2:
                        codons2[codon] = []
                    codons2[codon].append(count)
                    aa = translate(codon, table=translation_table)
                    if aa not in aa2:
                        aa2[aa] = 0
                    aa2[aa] = aa2[aa] + count
                break

    c1_keys = list(codons1.keys())
    c2_keys = list(codons2.keys())
    c_keys = list(set(c1_keys + c2_keys))
    c_keys.sort()
    print("Reporting %i codons" % len(c_keys))
    assert (
        len(c1_keys) == len(c2_keys) == len(c_keys)
    ), "Mismatching key lengths of codons"
    with open(args.out, "w") as fh:
        fh.write("#frame\tcodon_count\t%s\n" % ("\t".join(c_keys)))
        codons = {}
        for key, value in codons1.items():
            codons[key] = sum(value)
        codon_count = sum(codons.values())
        fh.write("2-counts\t%i" % codon_count)
        for key in c_keys:
            fh.write("\t%i" % (codons[key]))
        fh.write("\n")
        percent = 0
        fh.write("2-percent\t%i" % codon_count)
        for key in c_keys:
            p = codons[key] / codon_count * 100
            fh.write("\t%.28f" % (p))
            percent += p
        fh.write("\n")
        print("Found a total of %f percent for 2." % percent)

        codons = {}
        for key, value in codons2.items():
            codons[key] = sum(value)
        codon_count = sum(codons.values())
        fh.write("3-counts\t%i" % codon_count)
        for key in c_keys:
            fh.write("\t%i" % (codons[key]))
        fh.write("\n")
        percent = 0
        fh.write("3-percent\t%i" % codon_count)
        for key in c_keys:
            p = codons[key] / codon_count * 100
            fh.write("\t%.28f" % (p))
            percent += p
        fh.write("\n")
        print("Found a total of %f percent for 3." % percent)

    aa1_keys = list(aa1.keys())
    aa2_keys = list(aa2.keys())
    aa_keys = list(set(aa1_keys + aa2_keys))
    aa_keys.sort()
    print("Reporting %i amino acids" % len(aa_keys))
    assert (
        len(aa1_keys) == len(aa2_keys) == len(aa_keys)
    ), "Mismatching key lengths of amino acids"
    with open(args.aa_out, "w") as fh:
        fh.write("#frame\taa_count\t%s\n" % ("\t".join(aa_keys)))
        aas = aa1
        aa_count = sum(aas.values())
        fh.write("2-counts\t%i" % aa_count)
        for key in aa_keys:
            fh.write("\t%i" % (aas[key]))
        fh.write("\n")
        percent = 0
        fh.write("2-percent\t%i" % aa_count)
        for key in aa_keys:
            p = aas[key] / aa_count * 100
            fh.write("\t%.28f" % (p))
            percent += p
        fh.write("\n")
        print("Found a total of %f percent for 2-aa." % percent)

        aas = aa2
        aa_count = sum(aas.values())
        fh.write("3-counts\t%i" % aa_count)
        for key in aa_keys:
            fh.write("\t%i" % (aas[key]))
        fh.write("\n")
        percent = 0
        fh.write("3-percent\t%i" % aa_count)
        for key in aa_keys:
            p = aas[key] / aa_count * 100
            fh.write("\t%.28f" % (p))
            percent += p
        fh.write("\n")
        print("Found a total of %f percent for 3-aa." % percent)


if __name__ == "__main__":
    __main__()



#!/usr/bin/env python3

import argparse

from importlib.metadata import version
from sys import exit, stderr

from Bio.Seq import reverse_complement, translate

from twobitreader import TwoBitFile

FA_SEP = "|"

log_wtr = stderr


def get_bed_from_line(line, log_wtr=log_wtr):
    fields = line.rstrip("\r\n").split("\t")
    if len(fields) < 12:
        raise Exception("BADBED: Not a proper 12 column BED line (%s)." % line)
    (
        chrom,
        chromStart,
        chromEnd,
        name,
        score,
        strand,
        thickStart,
        thickEnd,
        itemRgb,
        blockCount,
        blockSizes,
        blockStarts,
    ) = fields[0:12]
    bed = BedInterval(
        chrom=chrom,
        chromStart=chromStart,
        chromEnd=chromEnd,
        name=name,
        score=score,
        strand=strand,
        thickStart=thickStart,
        thickEnd=thickEnd,
        itemRgb=itemRgb,
        blockCount=blockCount,
        blockSizes=blockSizes,
        blockStarts=blockStarts,
        log_wtr=log_wtr,
    )
    return bed


def make_int_list(obj, sep=","):
    if obj is None:
        return None
    if isinstance(obj, list):
        return [int(x) for x in obj]
    return [int(x) for x in obj.rstrip(sep).split(sep)]


def get_all_canonical_sequences(input_rdr, twobit, log_wtr=log_wtr):
    tell = input_rdr.tell()
    rval = []
    i = 0
    for i, bedline in enumerate(input_rdr, start=1):
        try:
            bed = get_bed_from_line(bedline)
            regions, cds = bed.get_cds(twobit)
            if cds and cds not in rval:
                rval.append(cds)
        except Exception as e:
            print("Cannot determine CDS for exclusion: %s" % str(e), file=log_wtr)
    print(
        "Excluding %i unique canonical coding sequences from %i BED lines."
        % (len(rval), i),
        file=log_wtr,
    )
    input_rdr.seek(tell)
    return rval


class BedInterval(object):
    def __init__(
        self,
        chrom=None,
        chromStart=None,
        chromEnd=None,
        name=None,
        score=None,
        strand=None,
        thickStart=None,
        thickEnd=None,
        itemRgb=None,
        blockCount=None,
        blockSizes=None,
        blockStarts=None,
        log_wtr=log_wtr,
    ):
        self.chrom = chrom
        self.chromStart = int(chromStart)
        self.chromEnd = int(chromEnd)
        self.name = name
        self.score = int(score) if score is not None else 0
        self.strand = "-" if str(strand).startswith("-") else "+"
        self.thickStart = int(thickStart) if thickStart else self.chromStart
        self.thickEnd = int(thickEnd) if thickEnd else self.chromEnd
        self.itemRgb = str(itemRgb) if itemRgb is not None else r"100,100,100"
        self.blockCount = int(blockCount)
        self.blockSizes = make_int_list(blockSizes)
        self.blockStarts = make_int_list(blockStarts)
        self.log_wtr = log_wtr

    def __str__(self):
        return "%s\t%d\t%d\t%s\t%d\t%s\t%d\t%d\t%s\t%d\t%s\t%s" % (
            self.chrom,
            self.chromStart,
            self.chromEnd,
            self.name,
            self.score,
            self.strand,
            self.thickStart,
            self.thickEnd,
            str(self.itemRgb),
            self.blockCount,
            ",".join([str(x) for x in self.blockSizes]),
            ",".join([str(x) for x in self.blockStarts]),
        )

    def get_cds_bed(self):
        trimmed = self.trim_by_region(self.thickStart, self.thickEnd)
        return trimmed

    def trim_by_region(self, region_start, region_end):
        starts = []
        ends = []
        exon_starts = [x + self.chromStart for x in self.blockStarts]
        exon_ends = [x + y for x, y in zip(exon_starts, self.blockSizes)]

        for start, end in zip(exon_starts, exon_ends):
            start = max(start, region_start)
            end = min(end, region_end)
            if start < end:
                starts.append(start)
                ends.append(end)

        thick_start = max(region_start, self.thickStart)
        thick_end = min(region_end, self.thickEnd)

        block_starts = [x - region_start for x in starts]
        block_sizes = [x - y for x, y in zip(ends, starts)]

        return BedInterval(
            chrom=self.chrom,
            chromStart=region_start,
            chromEnd=region_end,
            name=self.name,
            score=self.score,
            strand=self.strand,
            thickStart=thick_start,
            thickEnd=thick_end,
            itemRgb=self.itemRgb,
            blockCount=len(block_sizes),
            blockSizes=block_sizes,
            blockStarts=block_starts,
            log_wtr=self.log_wtr,
        )

    def trim_by_cds_offsets(self, start_offset, end_offset):
        cds_bed = self.get_cds_bed()
        if cds_bed.strand != "+":
            t = start_offset
            start_offset = end_offset
            end_offset = t
        starts = []
        sizes = []
        for start, size in zip(cds_bed.blockStarts, cds_bed.blockSizes):
            if start_offset:
                size = size - start_offset
                if size <= 0:
                    start_offset = abs(size)
                    continue
                start = start + start_offset
                start_offset = 0
            starts.append(start)
            sizes.append(size)
        starts2 = []
        sizes2 = []
        starts.reverse()
        sizes.reverse()
        for start, size in zip(starts, sizes):
            if end_offset:
                size = size - end_offset
                if size <= 0:
                    end_offset = abs(size)
                    continue
                end_offset = 0
            starts2.append(start)
            sizes2.append(size)
        starts2.reverse()
        sizes2.reverse()
        cds_bed.blockSizes = sizes2
        cds_bed.blockStarts = starts2
        cds_bed.blockCount = len(sizes2)

        start = cds_bed.blockStarts[0] + cds_bed.chromStart
        end = cds_bed.blockStarts[-1] + cds_bed.chromStart + cds_bed.blockSizes[-1]
        return cds_bed.trim_by_region(start, end)

    def get_cdna_exon_regions(self):
        return [
            (self.chromStart + x, self.chromStart + x + y)
            for x, y in zip(self.blockStarts, self.blockSizes)
        ]

    def get_coding_exon_regions(self):
        starts = []
        ends = []
        exon_starts = [x + self.chromStart for x in self.blockStarts]
        exon_ends = [x + y for x, y in zip(exon_starts, self.blockSizes)]
        for start, end in zip(exon_starts, exon_ends):
            start = max(start, self.thickStart)
            end = min(end, self.thickEnd)
            if start < end:
                starts.append(start)
                ends.append(end)
        return [(x, y) for x, y in zip(starts, ends)]

    def get_sequence_for_region(self, chrom, start, end, twobit, strand="+"):
        seq = None
        if chrom in twobit and 0 <= start < end < len(twobit[chrom]):
            seq = twobit[chrom][start:end]
        contig = chrom[3:] if chrom.startswith("chr") else "chr%s" % chrom
        if contig in twobit and 0 <= start < end < len(twobit[contig]):
            seq = twobit[contig][start:end]
        if not seq:
            print(
                "Seq is None (%s): %s. %s %s %s %s"
                % (seq, str(self), chrom, start, end, strand),
                file=self.log_wtr,
            )
        elif strand != "+":
            seq = reverse_complement(seq)
        return seq

    def get_sequence_for_regions(self, chrom, regions, twobit, strand="+"):
        seq = [
            self.get_sequence_for_region(chrom, start, end, twobit, strand=strand)
            for start, end in regions
        ]
        if None in seq:
            return None  # most likely the chromosome does not exist in twobit
        if strand != "+":
            seq.reverse()
        return "".join(seq)

    def get_sequence(self, twobit):
        return self.get_sequence_for_region(
            self.chrom, self.chromStart, self.chromEnd, twobit, strand=self.strand
        )

    def get_cdna(self, twobit):
        regions = self.get_cdna_exon_regions()
        return (
            regions,
            self.get_sequence_for_regions(
                self.chrom, regions, twobit, strand=self.strand
            ),
        )

    def get_cds(self, twobit):
        regions = self.get_coding_exon_regions()
        return (
            regions,
            self.get_sequence_for_regions(
                self.chrom, regions, twobit, strand=self.strand
            ),
        )


def __main__():
    parser = argparse.ArgumentParser(
        description="Find Nested Alternate ORFs (nAlt-ORFs)"
    )
    parser.add_argument("--bed", default=None, help="BED input")
    parser.add_argument(
        "-t", "--twobit", default=None, help="Genome reference sequence in 2bit format"
    )
    parser.add_argument(
        "-f",
        "--peptide_fasta_out",
        default=None,
        help="Path to output peptide_fasta_out.fasta",
    )
    parser.add_argument(
        "-n",
        "--naltorfs_fasta_out",
        default=None,
        help="Path to output naltorfs_fasta_out.fasta",
    )
    parser.add_argument(
        "-o", "--cds_fasta_out", default=None, help="Path to output cds_fasta_out.fasta"
    )
    parser.add_argument(
        "-b", "--bed_out", default=None, help="Path to output probed.bed"
    )
    parser.add_argument("--log", default=None, help="Path to output log")
    parser.add_argument(
        "-m",
        "--min_length",
        type=int,
        default=10,
        help="Minimum length of protein translation to report",
    )
    parser.add_argument("-r", "--reference", default="", help="Genome Reference Name")
    parser.add_argument(
        "--unique_sequences",
        action="store_true",
        default=False,
        help="Only report the first unique occurrence of an alternate sequence",
    )
    parser.add_argument(
        "--no_canonical_cds",
        default=None,
        help="Do not report any alternate sequences that match a provided cannonical CDS in file",
    )
    parser.add_argument(
        "--translation_table",
        type=int,
        default=1,
        help="Translation table to use.",
    )
    parser.add_argument(
        "--version", action="store_true", help="Report version and exit"
    )
    args = parser.parse_args()

    if args.version:
        print("nAltORFs {}".format(version("naltorfs")))
        exit()

    input_rdr = open(args.bed, "r")
    peptide_fasta_wtr = open(args.peptide_fasta_out, "w")
    lno_fasta_wtr = open(args.naltorfs_fasta_out, "w")
    cds_fasta_wtr = open(args.cds_fasta_out, "w")

    if args.log:
        log_wtr = open(args.log, "w")
    else:
        log_wtr = stderr

    bed_wtr = open(args.bed_out, "w")

    twobit = TwoBitFile(args.twobit)
    print(twobit.sequence_sizes(), file=log_wtr)

    if args.no_canonical_cds:
        with open(args.no_canonical_cds, "r") as fh:
            unique_seqs = get_all_canonical_sequences(fh, twobit, log_wtr)
    else:
        unique_seqs = []

    def non_canonical(seq):
        if seq not in unique_seqs:
            return True
        return False

    def unique_sequence(seq):
        if seq not in unique_seqs:
            unique_seqs.append(seq)
            return True
        return False

    def is_always_true(seq):
        return True

    is_unique = unique_sequence
    if not args.unique_sequences:
        if args.no_canonical_cds:
            is_unique = non_canonical
        else:
            is_unique = is_always_true

    def write_outputs(tbed, accession, peptide, nucseq, cds_bed, cds):
        probed = "%s\t%s\t%s\t%s%s" % (
            accession,
            peptide,
            "unique",
            args.reference,
            "\t." * 9,
        )
        if bed_wtr:
            bed_wtr.write("%s\t%s\n" % (str(tbed), probed))
            bed_wtr.flush()
        location = "chromosome:%s:%s:%s:%s:%s" % (
            args.reference,
            tbed.chrom,
            tbed.thickStart,
            tbed.thickEnd,
            tbed.strand,
        )
        fa_desc = "%s%s" % (FA_SEP, location)
        fa_id = ">%s%s\n" % (tbed.name, fa_desc)
        peptide_fasta_wtr.write(fa_id)
        peptide_fasta_wtr.write(peptide)
        peptide_fasta_wtr.write("\n")
        peptide_fasta_wtr.flush()

        fa_desc = "%s%s" % (FA_SEP, location)
        fa_id = ">%s%s\n" % (tbed.name, fa_desc)
        lno_fasta_wtr.write(fa_id)
        lno_fasta_wtr.write(nucseq)
        lno_fasta_wtr.write("\n")
        lno_fasta_wtr.flush()

        # fa_desc = '%s%s' % (args.fa_sep, location)
        fa_desc = ""
        fa_id = ">%s%s\n" % (cds_bed.name, fa_desc)
        cds_fasta_wtr.write(fa_id)
        cds_fasta_wtr.write(cds)
        cds_fasta_wtr.write("\n")
        cds_fasta_wtr.flush()

    def find_lno_orf(bed):
        translate_count = 0
        transcript_id = bed.name
        cdna_regions, cdna_seq = bed.get_cdna(twobit)
        cds_regions, cds_seq = bed.get_cds(twobit)

        if cdna_seq is None:
            print("cdna not found for BED: %s" % str(bed), file=log_wtr)
            return 0

        if cds_seq is None:
            print("cds not found for BED: %s" % str(bed), file=log_wtr)
            return 0
        cds_len = len(cds_seq)
        if cds_len % 3 != 0:
            cds_seq = cds_seq[: -(cds_len % 3)]
            print(
                "cds length is not divisible by 3!!!: %d. %s" % (cds_len, str(bed)),
                file=log_wtr,
            )  # why does this happen? check several manually, the records may be listed as incomplete

        cds_bed = bed.get_cds_bed()
        alt_cds = cds_seq[1:]  # original frame is now third (2)
        alt_cds_upper = alt_cds.upper()

        cds_offset = 1
        if "ATG" in alt_cds_upper:
            atg_index = alt_cds_upper.find("ATG")
            atg_mod = atg_index % 3
            if atg_mod == 2:
                return 0  # is is original frame
            elif atg_mod == 0:
                frame = 2
            else:
                frame = 3

            cds_offset = cds_offset + atg_index
            alt_cds = alt_cds[atg_index:]
            alt_cds_mod = len(alt_cds) % 3
            alt_cds = alt_cds[:-alt_cds_mod]

            aaseq = translate(alt_cds, table=args.translation_table)
            assert aaseq.startswith("M"), Exception(
                "Translation did not start with methionine: %s --> %s" % alt_cds, aaseq
            )
            aa_end = aaseq.find("*")
            if aa_end < 0:
                aa_end = len(aaseq)
            aaseq = aaseq[:aa_end]
            len_aaseq = len(aaseq)
            if len_aaseq >= args.min_length:
                alt_cds_len = len_aaseq * 3
                alt_cds_end_trim = cds_len - alt_cds_len - cds_offset
                alt_cds = alt_cds[:alt_cds_len]
                tbed = cds_bed.trim_by_cds_offsets(cds_offset, alt_cds_end_trim)
                prot_acc = "%s_f%d_%d_%d" % (
                    transcript_id,
                    frame,
                    cds_offset,
                    alt_cds_end_trim,
                )  # tbed.chromStart, tbed.chromEnd)
                if is_unique(alt_cds):
                    translate_count += 1
                    write_outputs(tbed, prot_acc, aaseq, alt_cds, cds_bed, cds_seq)
                    return 1
        return 0

    if input_rdr:
        lno_count = 0
        transcript_count = 0
        for i, bedline in enumerate(input_rdr):
            try:
                bed = get_bed_from_line(bedline, log_wtr=log_wtr)
                if bed is None:
                    continue
                transcript_count += 1
                lno_count += find_lno_orf(bed)
            except Exception as e:
                print(
                    "BED format Error: line %d: %s\n%s" % (i, bedline, e), file=log_wtr
                )
        print(
            "transcripts: %d\tunique LNO ORFs: %d" % (transcript_count, lno_count),
            file=log_wtr,
        )


if __name__ == "__main__":
    __main__()
